#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
 Convert an MRS crawl predictions file into the *sem expected file format.  Two files are required:
 one containing the mrs predictions, the other containing annotations from auxilliary systems. When the
 -i flag is specified then the created file will inherit scope predictions from the auxilliary system,
 if there is no prediction in the MRS crawl file.
"""


from argparse import ArgumentParser
from starsem import Data


def tsdb_to_starsem_id(tsdb_id):
	tsdb_id = str(tsdb_id)
	chapter_id = int(tsdb_id[:2])

	if chapter_id >= 10 and chapter_id <= 11:
		title = 'wisteria'
		first_chapter = 10
	elif chapter_id >= 12 and chapter_id <= 25:
		title = 'baskervilles'
		first_chapter = 12
	elif chapter_id == 26:
		return "cardboard", str(int(tsdb_id[2:]))
	elif chapter_id >= 27 and chapter_id <= 28:
		title = 'circle'
		first_chapter = 27
	else:
		assert False, 'unknown chapter id ' + str(chapter_id)

	chapter = '%s%02d' % (title, chapter_id - first_chapter + 1)
	sentence = str(int(tsdb_id[2:]))

	return chapter, sentence


class MrsCrawlPredictions(dict):
	def __init__(self, path):
		super(MrsCrawlPredictions, self).__init__()
		item = list()
		ifile = open(path)
		for line in ifile:
			line = line.strip()
			if line:
				item.append(Token(line))
			else:
				if len(item) > 0:
					chapter, sentence = tsdb_to_starsem_id(item[0].tsdb_id)
					item_key = MrsCrawlPredictions.get_item_key(chapter, sentence)
					if item_key not in self:
						self[item_key] = dict()
					cue_ids = [token.token_id for token in item if token.in_cue]
					cue_key = MrsCrawlPredictions.get_cue_key(cue_ids[0], cue_ids[-1])
					self[item_key][cue_key] = item
				item = list()
		ifile.close()

	@staticmethod
	def get_item_key(chapter, sentence):
		return chapter + ' ' + str(sentence)

	@staticmethod
	def get_cue_key(first_cue_index, last_cue_index):
		return  str(first_cue_index) + '-' + str(last_cue_index)


class Token(object):
	def __init__(self, line):
		fields = line.split('\t')
		self.tsdb_id = fields[0]
		self.token_id = fields[1]
		self.word = fields[2]
		self.in_scope = fields[3] == '1'
		self.inherit = fields[3] == 'I'
		self.in_cue = fields[4] == '1'


def main():

	argparser = ArgumentParser()
	argparser.add_argument('-a', '--aux', required=True,
		help='path to file with cue predictions etc from auxilliary systems (*sem format)')
	argparser.add_argument('-m', '--mrs-crawl', required=True,
		help='path to file with scope predictions from MRS crawling')
	argparser.add_argument('-i', '--inherit', action='store_true',
		help='use the existing scopes when an item does not exist in the mrs-crawl file')
	args = argparser.parse_args()

	# load dataset using existing code (see starsem.py)
	data = Data(args.aux)

	# remove existing scope predictions if they won't inherited
	if not args.inherit:
		for sentence in data:
			for token in sentence:
				for annotation_id in range(sentence.num_annotations):
					token.annotations[annotation_id].scope = None

	# load MRS crawl predictions
	item_predictions = MrsCrawlPredictions(args.mrs_crawl)

	# set scope according to MRS crawl predictions
	for sentence in data:
		if sentence.num_annotations > 0:
			item_key = MrsCrawlPredictions.get_item_key(sentence.chapter, sentence.number)
			if item_key in item_predictions:
				annotation_predictions = item_predictions[item_key]
				for annotation_id in range(sentence.num_annotations):
					cues = sentence.get_cues(annotation_id)
					cue_id = MrsCrawlPredictions.get_cue_key(cues[0][0], cues[-1][-1])
					if cue_id in annotation_predictions:
						prediction = annotation_predictions[cue_id]
						for token_id, token in enumerate(sentence):
							word = token.word
							assert word == prediction[token_id].word
							if prediction[token_id].in_scope:
								scope = word
							else:
								cue = token.annotations[annotation_id].cue
							 	if cue and len(cue) < len(word):
									# this is an affixal cue - rest of token should be in scope
									cue_position = word.find(cue)
									if cue_position > 0:
										scope = word[0:cue_position]
									else:
										scope = word[cue_position + len(cue):]
								else:
									scope = None
							if not prediction[token_id].inherit:
								token.annotations[annotation_id].scope = scope

	print data.to_conll()


#     if not cue:
#         scope = word
#     elif len(cue) < len(word):
#         cue_position = word.find(cue)
#         if cue_position > 0:
#             scope = word[0:cue_position]
#         else:
#             scope = word[cue_position + len(cue):]
#     else:
#         scope = None

#     token.annotations[annotation_id].scope = scope

if __name__ == '__main__':
	main()
