#!/usr/bin/env python
# -*- coding: utf-8 -*-

from argparse import ArgumentParser
from starsem import Data, tsdb_id
import codecs
import sys

def find_annotation(g_sentence, g_annotation_id, s_sentence):
	cue_start = g_sentence.get_cues(g_annotation_id)[0][0]
	for annotation_id in range(s_sentence.num_annotations):
		if s_sentence.get_cues(annotation_id)[0][0] == cue_start:
			return annotation_id
	# else
	return None

class Result(object):

	TP = 'tp'
	FP = 'fp'
	FN = 'fn'

	def __init__(self, name):
		self.name = name
		self.tp = 0
		self.fp = 0
		self.fn = 0

	def register(self, outcome):
		if outcome is Result.TP:
			self.tp += 1
		elif outcome is Result.FP:
			self.fp += 1
		elif outcome is Result.FN:
			self.fn += 1
		elif outcome is None:
			pass
		else:
			assert False, 'outcome %s unknown' % outcome

	def precision(self):
		return self.tp / float(self.tp + self.fp)

	def recall(self):
		return self.tp / float(self.tp + self.fn)

	def fscore(self):
		precision = self.precision()
		recall = self.recall()
		return (2 * precision * recall) / (precision + recall)

	def report(self):
		total = self.tp + self.fp + self.fn
		return '%s:\n\ttotal=%d tp=%d fp=%d fn=%d\n\tpre=%.04f rec=%.04f f1=%.04f' % (self.name,  total, self.tp, self.fp, self.fn, self.precision(), self.recall(), self.fscore())



def get_outcome(g_sentence, g_annotation_id, s_sentence):
	assert g_sentence.chapter == s_sentence.chapter
	assert g_sentence.number == s_sentence.number
	assert g_sentence.num_annotations == s_sentence.num_annotations

	s_annotation_id = find_annotation(g_sentence, g_annotation_id, s_sentence)
	assert s_annotation_id is not None

	prediction = s_sentence.pretty(s_annotation_id)

	if g_sentence.get_scopes(g_annotation_id):
		is_correct = True

		for j, g_token in enumerate(g_sentence):
			s_token = s_sentence[j]
			if g_token.annotations[g_annotation_id].scope != s_token.annotations[s_annotation_id].scope:
				is_correct = False
				break

		if is_correct:
			return prediction, Result.TP
		else:
			return prediction, Result.FN

	elif s_sentence.get_scopes(s_annotation_id):
		return prediction, Result.FP
	else:
		return prediction, None

def get_token_oracle_outcomes(g_sentence, g_annotation_id, sentence_a, sentence_b):
	assert g_sentence.chapter == sentence_a.chapter
	assert g_sentence.number == sentence_a.number
	assert g_sentence.num_annotations == sentence_a.num_annotations
	assert g_sentence.chapter == sentence_b.chapter
	assert g_sentence.number == sentence_b.number
	assert g_sentence.num_annotations == sentence_b.num_annotations

	a_annotation_id = find_annotation(g_sentence, g_annotation_id, sentence_a)
	b_annotation_id = find_annotation(g_sentence, g_annotation_id, sentence_b)
	outcomes = list()
	for i in range(len(g_sentence)):
		g_scope = g_sentence[i].annotations[g_annotation_id].scope
		a_scope = sentence_a[i].annotations[a_annotation_id].scope
		b_scope = sentence_b[i].annotations[b_annotation_id].scope
		if g_scope:
			if a_scope == g_scope or b_scope == g_scope:
				outcomes.append(Result.TP)
				#print g_scope, a_scope, b_scope, outcomes[-1]

			else:
				outcomes.append(Result.FN)
				#print g_scope, a_scope, b_scope, outcomes[-1]
		elif a_scope or b_scope:
			outcomes.append(Result.FP)
			#print g_scope, a_scope, b_scope, outcomes[-1]
		#else:
			#print g_scope, a_scope, b_scope, '--'


	return outcomes

def report(ofile, identifier, gold, a_prediction, b_prediction):
	ofile.write(str(identifier) + '\n')
	ofile.write('[{0:s}] {1:s}\n'.format('G', gold.replace('|', '')))
	ofile.write('[{0:s}] {1:s}\n'.format('A', a_prediction.replace('|', '')))
	ofile.write('[{0:s}] {1:s}\n'.format('B', b_prediction.replace('|', '')))
	ofile.write('\n')

def token_f1(gold, gold_annotation_id, system, system_annotation_id):

	tp = 0
	fp = 0
	fn = 0
	for i in range(len(gold)):
		if gold[i].annotations[gold_annotation_id].scope:
			if gold[i].annotations[gold_annotation_id].scope == system[i].annotations[system_annotation_id].scope:
				tp += 1
			else:
				fn += 1
		elif system[i].annotations[system_annotation_id].scope:
			fp += 1

	precision = tp / float(tp + fp) if tp + fp > 0 else 0
	recall = tp / float(tp + fn) if tp + fn > 0 else 0
	return (2 * precision * recall) / (precision + recall) if precision + recall > 0 else 0


def main():

	argparser = ArgumentParser()
	argparser.add_argument('-g', '--gold', required=True,
		help='path to file with gold standard')
	argparser.add_argument('-a', '--system-a', required=True,
		help='path to file with predictions from system A')
	argparser.add_argument('-b', '--system-b', required=True,
		help='path to file with predictions from system B; oracle will use these if A is incorrect')
	args = argparser.parse_args()

	gold = Data(args.gold)
	system_a = Data(args.system_a)
	system_b = Data(args.system_b)
	oracle = Data(args.gold)

	# remove all scope predictions
	for sentence in oracle:
		for token in sentence:
			for annotation_id in range(sentence.num_annotations):
				token.annotations[annotation_id].scope = None

	a_count = 0
	b_count = 0
	tie_count = 0
	for i in range(len(gold)):

		assert len(gold[i]) == len(system_a[i]) and len(gold[i]) == len(system_b[i])
		assert gold[i].num_annotations == system_a[i].num_annotations and gold[i].num_annotations == system_b[i].num_annotations

		for annotation_id in range(gold[i].num_annotations):

			a_annotation_id = find_annotation(gold[i], annotation_id, system_a[i])
			a_annotation_f1 = token_f1(gold[i], annotation_id, system_a[i], a_annotation_id)

			b_annotation_id = find_annotation(gold[i], annotation_id, system_b[i])
			b_annotation_f1 = token_f1(gold[i], annotation_id, system_b[i], b_annotation_id)

			if a_annotation_f1 > b_annotation_f1:
				a_count += 1
			else:
				if a_annotation_f1 == b_annotation_f1:
					tie_count += 1
				else:
					b_count += 1

			for j in range(len(gold[i])):
				if a_annotation_f1 > b_annotation_f1:
					oracle[i][j].annotations[annotation_id].scope = system_a[i][j].annotations[a_annotation_id].scope
				else:
					oracle[i][j].annotations[annotation_id].scope = system_b[i][j].annotations[b_annotation_id].scope

	print oracle.to_conll()
	sys.stderr.write("%d A win, %d B win, %d tie (used B)\n"%(a_count,b_count,tie_count))

if __name__ == '__main__':
	main()
